import express from 'express';
import cors from 'cors';
import http from 'http';
/**
 * Create a minimal HTTP server that exposes SQLite data
 * @param db The SQLite database instance
 * @param port The port to run the server on (default: 3000)
 * @returns The HTTP server instance
 */
export function createHttpServer(db, port = 3000) {
    const app = express();
    app.use(cors());
    // Create server instance
    const httpServer = http.createServer(app);
    // Endpoint to get table or view data in CSV or JSON format
    app.get('/data/:name', (req, res) => {
        (async () => {
            try {
                const { name } = req.params;
                const format = req.query.format || 'csv';
                const limit = parseInt(req.query.limit) || 1000;
                // Execute the query with read-only mode
                const data = await db.executeQuery(`SELECT * FROM ${name} LIMIT ${limit}`, true);
                // Return data in requested format
                if (format === 'json') {
                    res.json(data);
                }
                else {
                    // Convert to CSV
                    res.setHeader('Content-Type', 'text/csv');
                    if (data.length === 0) {
                        return res.send('');
                    }
                    // Get headers from the first row
                    const headers = Object.keys(data[0]);
                    let csvContent = headers.join(',') + '\n';
                    // Add data rows
                    for (const row of data) {
                        const values = headers.map(header => {
                            const value = row[header];
                            if (value === null || value === undefined)
                                return '';
                            if (typeof value === 'string' && (value.includes(',') || value.includes('"') || value.includes('\n'))) {
                                return `"${value.replace(/"/g, '""')}"`;
                            }
                            return String(value);
                        });
                        csvContent += values.join(',') + '\n';
                    }
                    res.send(csvContent);
                }
            }
            catch (error) {
                const errorMessage = error instanceof Error ? error.message : String(error);
                res.status(500).json({ error: `Error getting data: ${errorMessage}` });
            }
        })();
    });
    // Endpoint to execute a query and return the results
    app.get('/query', (req, res) => {
        (async () => {
            try {
                const sql = req.query.sql;
                const format = req.query.format || 'csv';
                if (!sql) {
                    return res.status(400).json({ error: 'Missing SQL query' });
                }
                // Only allow SELECT queries
                if (!sql.trim().toLowerCase().startsWith('select')) {
                    return res.status(403).json({ error: 'Only SELECT queries are allowed' });
                }
                // Execute the query in read-only mode
                const result = await db.executeQuery(sql, true);
                // Return results in requested format
                if (format === 'json') {
                    res.json(result);
                }
                else {
                    // Convert to CSV
                    res.setHeader('Content-Type', 'text/csv');
                    if (result.length === 0) {
                        return res.send('');
                    }
                    // Get headers from the first row
                    const headers = Object.keys(result[0]);
                    let csvContent = headers.join(',') + '\n';
                    // Add data rows
                    for (const row of result) {
                        const values = headers.map(header => {
                            const value = row[header];
                            if (value === null || value === undefined)
                                return '';
                            if (typeof value === 'string' && (value.includes(',') || value.includes('"') || value.includes('\n'))) {
                                return `"${value.replace(/"/g, '""')}"`;
                            }
                            return String(value);
                        });
                        csvContent += values.join(',') + '\n';
                    }
                    res.send(csvContent);
                }
            }
            catch (error) {
                const errorMessage = error instanceof Error ? error.message : String(error);
                res.status(500).json({ error: `Error executing query: ${errorMessage}` });
            }
        })();
    });
    // Start the server
    httpServer.listen(port, () => {
        console.log(`HTTP server running at http://localhost:${port}`);
    });
    return httpServer;
}
