#!/usr/bin/env node
import { McpServer, ResourceTemplate } from '@modelcontextprotocol/sdk/server/mcp.js';
import { StdioServerTransport } from '@modelcontextprotocol/sdk/server/stdio.js';
import sqlite3 from 'sqlite3';
import { z } from 'zod';
import path from 'path';
import fs from 'fs/promises';
import { createHttpServer } from './http-server.js';
// Command line argument parsing
const args = process.argv.slice(2);
if (args.length !== 1) {
    console.error('Usage: mcp-server-sqlite-npx <database-path>');
    process.exit(1);
}
const dbPath = path.resolve(args[0]);
// Default HTTP port (can be overridden via environment variable)
const HTTP_PORT = process.env.MCP_HTTP_PORT ? parseInt(process.env.MCP_HTTP_PORT) : 3000;
// Schema definitions
const ExecuteQueryArgsSchema = z.object({
    query: z.string().describe('SQL query to execute'),
    read_only: z.boolean().optional().describe('If true, only SELECT queries will be allowed'),
});
/**
 * Wrapper for sqlite3.Database that bridges CommonJS and ESM modules.
 * This abstraction is necessary because:
 * 1. sqlite3 is a CommonJS module while we're using ESM (type: "module")
 * 2. The module interop requires careful handling of the Database import
 * 3. We need to promisify the callback-based API to work better with async/await
 */
class DatabaseWrapper {
    db;
    constructor(filename) {
        this.db = new sqlite3.Database(filename);
    }
    query(sql, params = []) {
        return new Promise((resolve, reject) => {
            this.db.all(sql, params, (err, rows) => {
                if (err)
                    reject(err);
                else
                    resolve(rows);
            });
        });
    }
    execute(sql, params = []) {
        return new Promise((resolve, reject) => {
            this.db.run(sql, params, function (err) {
                if (err)
                    reject(err);
                else
                    resolve([{ affectedRows: this.changes }]);
            });
        });
    }
}
class SqliteDatabase {
    db;
    constructor(dbPath) {
        this.db = new DatabaseWrapper(dbPath);
    }
    async query(sql, params = []) {
        return this.db.query(sql, params);
    }
    async listTables() {
        return this.query("SELECT name FROM sqlite_master WHERE type='table'");
    }
    async executeQuery(query, readOnly, params = []) {
        const trimmedQuery = query.trim().toUpperCase();
        // Validate query if read-only mode is enabled
        if (readOnly === true && !trimmedQuery.startsWith('SELECT')) {
            throw new Error('Only SELECT queries are allowed in read-only mode');
        }
        // For safety, prevent certain destructive operations
        if (trimmedQuery.startsWith('DROP ') ||
            trimmedQuery.includes('DELETE FROM') && !trimmedQuery.includes('WHERE')) {
            throw new Error('Potentially destructive query detected. Please add constraints or remove this safety check if intended.');
        }
        return this.query(query, params);
    }
}
// Export the SqliteDatabase class for other modules
export { SqliteDatabase };
// Server setup using the higher-level McpServer class
const server = new McpServer({
    name: 'sqlite-manager',
    version: '0.5.0',
});
const db = new SqliteDatabase(dbPath);
// Initialize the HTTP server
const httpServer = createHttpServer(db, HTTP_PORT);
// Add resources to expose database schema and table data
// Resource for database schema
server.resource('schema', 'sqlite://schema', async (uri) => {
    try {
        const tables = await db.listTables();
        const tableSchemas = await Promise.all(tables.map(async (table) => {
            const schema = await db.executeQuery(`PRAGMA table_info(${table.name})`, true);
            return {
                name: table.name,
                columns: schema
            };
        }));
        // Convert to CSV format
        let csvContent = 'table_name,column_name,type,notnull,default_value,pk\n';
        for (const table of tableSchemas) {
            for (const column of table.columns) {
                csvContent += `${table.name},${column.name},${column.type},${column.notnull},${column.dflt_value || ''},${column.pk}\n`;
            }
        }
        return {
            contents: [{
                    uri: uri.href,
                    text: csvContent,
                    mimeType: 'text/csv'
                }]
        };
    }
    catch (error) {
        const errorMessage = error instanceof Error ? error.message : String(error);
        throw new Error(`Error getting schema: ${errorMessage}`);
    }
});
// Dynamic resource for table data
server.resource('table-data', new ResourceTemplate('sqlite://tables/{tableName}', {
    list: async () => {
        try {
            // Get regular tables
            const tables = await db.listTables();
            // Get views
            const views = await db.executeQuery("SELECT name FROM sqlite_master WHERE type='view'", false);
            // Combine tables and views
            const resources = [
                ...tables.map((table) => ({
                    name: table.name,
                    uri: `sqlite://tables/${table.name}`,
                    description: `Data from the ${table.name} table`,
                    mimeType: 'text/csv'
                })),
                ...views.map((view) => ({
                    name: view.name,
                    uri: `sqlite://tables/${view.name}`,
                    description: `Data from the ${view.name} view`,
                    mimeType: 'text/csv'
                }))
            ];
            return { resources };
        }
        catch (error) {
            console.error('Error listing tables and views:', error);
            return { resources: [] };
        }
    }
}), async (uri, { tableName }) => {
    try {
        // Check if it's a table or view
        const entityType = await db.executeQuery("SELECT type FROM sqlite_master WHERE name = ? AND (type = 'table' OR type = 'view')", false, [tableName]);
        if (entityType.length === 0) {
            throw new Error(`Table or view '${tableName}' not found`);
        }
        const type = entityType[0].type;
        // Get a preview of the data (first 10 rows)
        const data = await db.executeQuery(`SELECT * FROM ${tableName} LIMIT 10`, true);
        // Convert to CSV format
        if (data.length === 0) {
            return {
                contents: [{
                        uri: uri.href,
                        text: `No data found in ${type} ${tableName}`,
                        mimeType: 'text/plain'
                    }]
            };
        }
        // Get headers from the first row
        const headers = Object.keys(data[0]);
        let csvContent = headers.join(',') + '\n';
        // Add data rows
        for (const row of data) {
            const values = headers.map(header => {
                const value = row[header];
                // Handle special cases like null, undefined, or values with commas
                if (value === null || value === undefined)
                    return '';
                if (typeof value === 'string' && (value.includes(',') || value.includes('"') || value.includes('\n'))) {
                    return `"${value.replace(/"/g, '""')}"`;
                }
                return String(value);
            });
            csvContent += values.join(',') + '\n';
        }
        // Include the HTTP URL to access the full dataset
        const httpUrl = `http://localhost:${HTTP_PORT}/data/${tableName}`;
        const fullContent = `# Preview (first 10 rows):\n${csvContent}\n\n# Full data available at:\n${httpUrl}\n\n# Access formats:\n- CSV: ${httpUrl}\n- JSON: ${httpUrl}?format=json`;
        return {
            contents: [{
                    uri: uri.href,
                    text: fullContent,
                    mimeType: 'text/markdown'
                }]
        };
    }
    catch (error) {
        const errorMessage = error instanceof Error ? error.message : String(error);
        throw new Error(`Error getting table data: ${errorMessage}`);
    }
});
// Add prompts for common SQL operations
server.prompt('select-data', {
    table: z.string(),
    columns: z.string().optional(),
    where: z.string().optional(),
    limit: z.string().optional()
}, (args) => {
    const { table, columns = '*', where, limit } = args;
    let query = `SELECT ${columns} FROM ${table}`;
    if (where) {
        query += ` WHERE ${where}`;
    }
    if (limit) {
        query += ` LIMIT ${limit}`;
    }
    return {
        messages: [{
                role: 'user',
                content: {
                    type: 'text',
                    text: `Please execute the following SQL query and analyze the results:\n\n\`\`\`sql\n${query}\n\`\`\``
                }
            }]
    };
});
server.prompt('analyze-table', {
    table: z.string()
}, (args) => {
    const { table } = args;
    return {
        messages: [{
                role: 'user',
                content: {
                    type: 'text',
                    text: `Please analyze the structure and data of the "${table}" table. First, examine the schema, then look at sample data, and provide insights about the table's purpose, data types, and any patterns you observe.`
                }
            }]
    };
});
// Add tools to the server using the simplified API
server.tool('execute_query', {
    query: z.string(),
    read_only: z.boolean().optional()
}, async ({ query, read_only }) => {
    try {
        const result = await db.executeQuery(query, read_only);
        // Convert to CSV format
        if (!Array.isArray(result) || result.length === 0) {
            return {
                content: [{ type: 'text', text: 'Query executed successfully. No results returned or non-SELECT query.' }]
            };
        }
        // Get headers from the first row
        const headers = Object.keys(result[0]);
        let csvContent = headers.join(',') + '\n';
        // Add data rows
        for (const row of result) {
            const values = headers.map(header => {
                const value = row[header];
                // Handle special cases like null, undefined, or values with commas
                if (value === null || value === undefined)
                    return '';
                if (typeof value === 'string' && (value.includes(',') || value.includes('"') || value.includes('\n'))) {
                    return `"${value.replace(/"/g, '""')}"`;
                }
                return String(value);
            });
            csvContent += values.join(',') + '\n';
        }
        return {
            content: [{ type: 'text', text: csvContent }]
        };
    }
    catch (error) {
        const errorMessage = error instanceof Error ? error.message : String(error);
        return {
            content: [{ type: 'text', text: `Error executing query: ${errorMessage}` }]
        };
    }
});
server.tool('list_tables', {}, async () => {
    try {
        const tables = await db.listTables();
        return {
            content: [{ type: 'text', text: JSON.stringify(tables, null, 2) }]
        };
    }
    catch (error) {
        const errorMessage = error instanceof Error ? error.message : String(error);
        return {
            content: [{ type: 'text', text: `Error: ${errorMessage}` }],
            isError: true
        };
    }
});
// Add a tool to save the entire database file
server.tool('save_database', {
    filepath: z.string().optional().describe('Path to save the database file to (defaults to original path)')
}, async ({ filepath }) => {
    try {
        // If no filepath is provided, use the original database path
        const targetPath = filepath || dbPath;
        if (targetPath === dbPath) {
            // If saving to the same file, we need to create a temporary copy first
            const tempPath = `${dbPath}.tmp`;
            await fs.copyFile(dbPath, tempPath);
            await fs.rename(tempPath, dbPath);
        }
        else {
            // Copy to a different location
            await fs.copyFile(dbPath, targetPath);
        }
        return {
            content: [{ type: 'text', text: `Successfully saved database to ${targetPath}` }]
        };
    }
    catch (error) {
        const errorMessage = error instanceof Error ? error.message : String(error);
        return {
            content: [{ type: 'text', text: `Error saving database: ${errorMessage}` }],
            isError: true
        };
    }
});
// Add a tool for SQL queries with HTTP results
server.tool('execute_query_with_url', {
    query: z.string().describe('SQL query to execute (must be a SELECT query)'),
    description: z.string().optional().describe('Optional description of the query purpose')
}, async ({ query, description }) => {
    try {
        // Only allow SELECT queries for security
        if (!query.trim().toLowerCase().startsWith('select')) {
            return {
                content: [{ type: 'text', text: 'Error: Only SELECT queries are allowed with this tool.' }]
            };
        }
        // Execute the query
        const result = await db.executeQuery(query, true);
        if (!Array.isArray(result) || result.length === 0) {
            return {
                content: [{ type: 'text', text: 'Query executed successfully but returned no results.' }]
            };
        }
        // Create a simple preview (first 5 rows)
        const preview = result.slice(0, 5);
        // Encode the query for use in a URL
        const encodedQuery = encodeURIComponent(query);
        const queryUrl = `http://localhost:${HTTP_PORT}/query?sql=${encodedQuery}`;
        // Return the result info
        const queryDesc = description ? description : 'Query result';
        return {
            content: [{
                    type: 'text',
                    text: `${queryDesc}\n\nPreview (first 5 rows):\n${JSON.stringify(preview, null, 2)}\n\nTotal rows: ${result.length}\n\nFull results available at:\n- CSV: ${queryUrl}\n- JSON: ${queryUrl}&format=json`
                }]
        };
    }
    catch (error) {
        const errorMessage = error instanceof Error ? error.message : String(error);
        return {
            content: [{ type: 'text', text: `Error executing query: ${errorMessage}` }]
        };
    }
});
// Add a tool to create a view
server.tool('create_view', {
    name: z.string().describe('Name for the view'),
    query: z.string().describe('SELECT query that defines the view'),
    description: z.string().optional().describe('Optional description of the view purpose')
}, async ({ name, query, description }) => {
    try {
        // Validate that the query starts with SELECT
        if (!query.trim().toLowerCase().startsWith('select')) {
            return {
                content: [{ type: 'text', text: 'Error: View definition must be a SELECT query.' }]
            };
        }
        // Create the view
        const createViewQuery = `CREATE VIEW IF NOT EXISTS ${name} AS ${query}`;
        await db.executeQuery(createViewQuery, false);
        // If description is provided, store it in a special table
        if (description) {
            // First, check if our metadata table exists, if not create it
            await db.executeQuery(`
          CREATE TABLE IF NOT EXISTS _mcp_metadata (
            object_type TEXT,
            object_name TEXT,
            description TEXT,
            created_at TEXT,
            PRIMARY KEY (object_type, object_name)
          )
        `, false);
            // Store the description
            await db.executeQuery(`
          INSERT OR REPLACE INTO _mcp_metadata (object_type, object_name, description, created_at)
          VALUES ('view', ?, ?, datetime('now'))
        `, false, [name, description]);
        }
        // Generate the HTTP URL for the view
        const httpUrl = `http://localhost:${HTTP_PORT}/data/${name}`;
        return {
            content: [{
                    type: 'text',
                    text: `View '${name}' created successfully.\n\nView data available at:\n- CSV: ${httpUrl}\n- JSON: ${httpUrl}?format=json`
                }]
        };
    }
    catch (error) {
        const errorMessage = error instanceof Error ? error.message : String(error);
        return {
            content: [{ type: 'text', text: `Error creating view: ${errorMessage}` }]
        };
    }
});
// Start server
async function runServer() {
    const transport = new StdioServerTransport();
    await server.connect(transport);
    // Use console.error to show error output.
    // console.log results in JSon exception.
    console.error('SQLite MCP Server running on stdio');
    console.error('Database path:', dbPath);
}
runServer().catch(error => {
    console.error('Fatal error running server:', error);
    process.exit(1);
});
